/**
  Energy Management with Opta™ - Application Note
  Name: energy_management
  Purpose: Based on the power profile and state, manage the power distribution for devices of interest.

  @author Arduino
  @version 1.1 10/05/2023 
*/

#include "stm32h7xx_ll_gpio.h"
#include "thingProperties.h"
#include <ArduinoModbus.h>
#include <ArduinoRS485.h>
#include <Scheduler.h>

#define F7M24 0x21
#define pause_trigger 15

// Energy Meter Parameters
float V_actual, V_avg, V_max, V_min;
float A_actual, A_avg, A_max, A_min;
float W_actual, W_avg, W_max, W_min;
float Var_actual;
float Va_actual, Va_avg, Va_max, Va_min;
float Wh_packet, Varh_packet, Wh_Abs_packet;

typedef struct uPWR_STRUCT {
  float uV_code;
  float uW_code;
  float uWh_code;
} PWR_STRUCT;
PWR_STRUCT user_profile;

/**
  ======================= IMPORTANT =======================
  Before using the system, the following parameters MUST be defined:
 
  - operation_safety_margin: Multiplier indicating safety margin. E.g., 1.1 implies a 10% headroom.
  - estimated_max_power: Expected maximum power (in Watts) under which the network will operate.
  - estimated_max_energy: Expected maximum active energy limit (in Wh) for the network.
  - Device_1_Limiter & Device_2_Limiter: Upper limit current (in Amperes) and power (in Watts) for Device 1 & 2 respectively.
  - Device_1_CompRef & Device_2_CompRef: Energy meter parameters required for accurate system operation.
*/
float operation_safety_margin = 1.1;
float estimated_max_power = 220;
float estimated_max_energy = 2880;
float Device_1_Limiter = 1;
float Device_2_Limiter = 100;
float Device_1_CompRef = A_actual;
float Device_2_CompRef = W_avg;
/*
  It is CRUCIAL to set these parameters accurately to ensure safe and correct operation of the system.
  =========================================================
*/

constexpr auto baudrate { 19200 };

// Calculate preDelay and postDelay in microseconds as per Modbus RTU Specification
// MODBUS over serial line specification and implementation guide V1.02
// Paragraph 2.5.1.1 MODBUS Message RTU Framing
// https://modbus.org/docs/Modbus_over_serial_line_V1_02.pdf
constexpr auto bitduration { 1.f / baudrate };
constexpr auto preDelayBR { bitduration * 9.6f * 3.5f * 1e6 };
constexpr auto postDelayBR { bitduration * 9.6f * 3.5f * 1e6 };

int start_time = 0;
int rs485_counter = 0;
int counter = 0;

// Device state flag
int Device_1_f = 0;
int Device_2_f = 0;

/*************************************
* Energy Meter related routines
*************************************/

/**
  Control the relay output based on the user (consumption) profile input and configured power/energy target.

  @param desired_target Desired resource required to run the connected device on the relay.
  @param req_target Minimum resource required to run the connected device on the relay.
  @param relayTarget Relay to activate or deactivate.
  @return Returns 0 or 1, representing HIGH state for 1 and LOW state for 0.
*/
uint8_t relay_Trigger(int desired_target, int req_target, pin_size_t relayTarget) {
  bool isStable = (desired_target >= req_target) && (desired_target > (req_target * user_profile.uV_code));
  digitalWrite(relayTarget, isStable ? HIGH : LOW);
  Serial.print(F("Energy Manager: "));
  Serial.print(isStable ? F("Stable operation margin") : F("Unstable / possible overload"));
  Serial.print(F(" - Turning "));
  Serial.print(isStable ? F("ON: ") : F("OFF: "));
  Serial.println(relayTarget);

  return isStable ? 1 : 0;
}

/**
  Initial user profile setup.

  @param init_OperMargin  System operation margin for power budget represented in percentage.
  @param init_Watt  User defined Wattage limit for the system.
  @param init_WhCon User defined Energy consumption limit for the system.
*/
void consumption_profile(float init_OperMargin, float init_Watt, float init_WhCon){
  uOperMargin = init_OperMargin;
  user_profile.uV_code = uOperMargin;

  uWatt = init_Watt;
  user_profile.uW_code = uWatt;

  uWhCon = init_WhCon;
  user_profile.uWh_code = uWhCon;
}

/**
  Simple consumption profile value updater.
*/
void consumption_var_container(){
  uOperMargin = user_profile.uV_code;
  uWatt = user_profile.uW_code;
  uWhCon = user_profile.uWh_code;
}

/**
  Handler functions for energy_distro_ctrl() funtion
*/
bool handleDevice(float threshold, float parameter_reference_data, int pin) {
  bool deviceFlag = relay_Trigger(threshold, parameter_reference_data, pin);
  if (!deviceFlag) {
      if (relay_Trigger(threshold, W_actual, pin)) {
          Serial.println(deviceFlag == Device_1_f ? F("Energy Manager: Secondary condition pass, may be unstable") : F("Energy Manager: Secondary Power condition pass"));
      } else {
          Serial.println(F("Energy Manager: Insufficient resource"));
          digitalWrite(D0, LOW);
      }
  } else {
      Serial.println(F("Energy Manager: Conditions green"));
  }
  return deviceFlag;
}

/**
  Monitors and uses the user defined profile and retrieved information from Energy meter to manage connected devices of interest.

  @param Wh_packet Retrieved energy information from Energy meter in unit of [Wh].
  @param Device_#_f Device flag controlled by relay_Trigger() function. # specifies device number or designation.
  @param directOverride1 Direct override flag for Device #1 controlled via Cloud.
  @param W_avg Average power information retrieved from Energy meter. 
  @param Device_X_Limiter User defined upper limit value for selected device parameter.
  @param Device_X_CompRef User selected Energy Meter parameter reading to be compared as a reference for device actuation and system behavior.
*/
void energy_distro_ctrl() {
  if (Wh_packet != 0) {
      uWhOpta = Wh_packet;
  } else {
      Serial.println(F("Energy Manager: Energy information recollection stand-by"));
      return;
  }

  if ((Wh_packet * user_profile.uV_code) <= user_profile.uWh_code) {
      Device_1_f = handleDevice(Device_1_Limiter, Device_1_CompRef, D0);
      Device_2_f = handleDevice(Device_2_Limiter, Device_2_CompRef, D1);
  } else {
      digitalWrite(D0, LOW);
      digitalWrite(D1, LOW);
      Serial.println(F("Energy Manager: Energy consumption is high! - Warning"));
  }

  if (directOverride1) {
      Serial.println(F("Energy Manager: Direct Override Request Received"));
      digitalWrite(D0, HIGH);
  }

  if ((W_avg * user_profile.uV_code) > user_profile.uW_code) {
      Serial.println(F("Energy Manager: Average Power is above profile limit! - Warning"));
      Serial.println(W_avg * user_profile.uV_code);
  }
}

/**
  Displays electrical, power, and energy consumption information retrieved from the energy meter.

  @param V_XXX Voltage information from Energy meter categorized in actual, average, maximum, and minimum value.
  @param A_XXX Current information from Energy meter categorized in actual, average, maximum, and minimum value.
  @param W_XXX Power information from Energy meter categorized in actual, average, maximum, and minimum value.
  @param Var_XXX Reactive power total information from Energy meter categorized in actual.
  @param Va_XXX Apparent power total information from Energy meter categorized in actual, average, maximum, and minimum value.
  @param Wh_packet Power based energy information from Energy meter.
  @param Varh_packet Reactive power total based energy information from Energy meter.
  @param Wh_Abs_packet Absolute Active total based energy information from Energy meter.
*/
struct DataGroup {
  const char* header;
  float* data;
  const char** units;
  uint8_t size;
};

const char* energyUnits[] = { "Wh", "Varh", "Wh" };
const char* actUnits[] = { "V", "A", "W", "Var", "VA" };
const char* avgUnits[] = { "V", "A", "W", "VA" };
const char* maxUnits[] = { "V", "A", "W", "VA" };
const char* minUnits[] = { "V", "A", "W", "VA" };
const char* cloudParamUnits[] = { "u", "W", "Wh" };

void modbus_com_monitor(){
  float energyData[] = { Wh_packet, Varh_packet, Wh_Abs_packet };
  float actData[] = { V_actual, A_actual, W_actual, Var_actual, Va_actual };
  float avgData[] = { V_avg, A_avg, W_avg, Va_avg };
  float maxData[] = { V_max, A_max, W_max, Va_max };
  float minData[] = { V_min, A_min, W_min, Va_min };
  float cloudParamData[] = { user_profile.uV_code, user_profile.uW_code, user_profile.uWh_code };

  DataGroup groups[] = {
    { "||| Energy  |||||||||||||||||||||", energyData, energyUnits, sizeof(energyData) / sizeof(energyData[0]) },
    { "||| Act. Data  ||||||||||||||||||", actData, actUnits, sizeof(actData) / sizeof(actData[0]) },
    { "||| AVG Data  |||||||||||||||||||", avgData, avgUnits, sizeof(avgData) / sizeof(avgData[0]) },
    { "||| Max. Data  ||||||||||||||||||", maxData, maxUnits, sizeof(maxData) / sizeof(maxData[0]) },
    { "||| Min. Data  ||||||||||||||||||", minData, minUnits, sizeof(minData) / sizeof(minData[0]) },
    { "||| Cloud Parameter Data  |||||||", cloudParamData, cloudParamUnits, sizeof(cloudParamData) / sizeof(cloudParamData[0]) }
  };

  Serial.println(); // Empty println for formating reasons
  for(uint8_t i = 0; i < sizeof(groups) / sizeof(groups[0]); i++) {
    Serial.println(F(groups[i].header));
    for(uint8_t j = 0; j < groups[i].size; j++) {
      Serial.print(groups[i].data[j]);
      Serial.print(" ");
      Serial.println(groups[i].units[j]);
    }
  }
  Serial.println(); // Empty println for formating reasons
}

/**
  Requests and retrieves actual electrical, and power information from Energy meter over Modbus RTU protocol.
*/
void modbus_com_actual(){
  // Actual Measurements
  // Voltage (V)
  V_actual = getT5(F7M24, 30107);

  // Current (A)
  A_actual = getT5(F7M24, 30126);

  // Active Power Total - Pt (W)
  W_actual = getT6(F7M24, 30140);
  
  // Reactive Power Total - Qt (var) (IEEE 754)
  Var_actual = getT_Float(F7M24, 32544);

  // Apparent Power Total - St (VA)
  Va_actual = getT5(F7M24, 30156);
}

/**
  Requests and retrieves average electrical, and power information from Energy meter over Modbus RTU protocol.
*/
void modbus_com_avg(){
  // Average Measurements
  // Voltage (V)
  V_avg = getT5(F7M24, 35507);

  // Current (A)
  A_avg = getT5(F7M24, 35526);

  // Active Power Total - Pt (W)
  W_avg = getT6(F7M24, 35540);

  // Apparent Power Total - St (VA)
  Va_avg = getT5(F7M24, 35556);
}

/**
  Requests and retrieves maximum electrical,and power information from Energy meter over Modbus RTU protocol.
*/
void modbus_com_max(){
  // Maximum Measurements
  // Voltage (V)
  V_max = getT5(F7M24, 35607);

  // Current (A)
  A_max = getT5(F7M24, 35626);

  // Active Power Total - Pt (W)
  W_max = getT6(F7M24, 35640);

  // Apparent Power Total - St (VA)
  Va_max = getT5(F7M24, 35656);
}

/**
  Requests and retrieves minimum electrical, and power information from Energy meter over Modbus RTU protocol.
*/
void modbus_com_min(){
  // Minimum Measurements
  // Voltage (V)
  V_min = getT5(F7M24, 35707);

  // Current (A)
  A_min = getT5(F7M24, 35726);

  // Active Power Total - Pt (W)
  W_min = getT6(F7M24, 35740);

  // Apparent Power Total - St (VA)
  Va_min = getT5(F7M24, 35756);
}

/**
  Requests and retrieves energy information from Energy meter over Modbus RTU protocol.
*/
void modbus_com_energy(){
  // Energy
  // Energy (Wh) - n1
  Wh_packet = getT_Float(F7M24, 32752);

  // Energy (varh) - n4
  Varh_packet = getT_Float(F7M24, 32758);

  // Total Absolute Active Energy (Wh)
  Wh_Abs_packet = getT_Float(F7M24, 32760);
}

/*************************************
* LED PLC Switches
*************************************/

/**
  Defaulting LED states to low when called. 
*/
void plc_led_off() {
  digitalWrite(LEDG, LOW);
  digitalWrite(LEDR, LOW);
  digitalWrite(LEDB, LOW);
  digitalWrite(PIN_SPI_MISO, LOW);
  digitalWrite(PIN_SPI_MOSI, LOW);
  digitalWrite(PIN_SPI_SCK, LOW);
  digitalWrite(PIN_SPI_SS, LOW);
}

/**
  Sets up the LEDs as outputs and calls the plc_led_off() function.
*/
void plc_led_Setup(){
  pinMode(LEDG, OUTPUT);
  pinMode(LEDR, OUTPUT);
  pinMode(LEDB, OUTPUT);
  pinMode(PIN_SPI_MISO, OUTPUT);
  pinMode(PIN_SPI_MOSI, OUTPUT);
  pinMode(PIN_SPI_SCK, OUTPUT);
  pinMode(PIN_SPI_SS, OUTPUT);

  plc_led_off();
}

/*************************************
* Digital Port related tasks
*************************************/

/**
  Sets up the desired output relay to control the pump.
*/
void digitalIO_Setup(){
  pinMode(D0, OUTPUT);
  pinMode(D1, OUTPUT);
}

/*************************************
* Analog Port related tasks
*************************************/

/**
  Sets up analog ports with 12 bit resolution.
*/
void analogIO_Setup(){
  analogReadResolution(12);

  start_time = millis();
  SCB_DisableDCache();
  Serial.println("Start");
}

/*************************************
* RTU related tasks
*************************************/

/**
  Sets up Modbus RTU protocol configuration.
*/
void RTU_Setup(){
  Serial.println("Energy Management - Modbus RTU Client");

  RS485.setDelays(preDelayBR, postDelayBR);

  // start the Modbus RTU client
  // 7M.24 Energy meter specifies 19200 of default baudrate and 8N2 frame
  if (!ModbusRTUClient.begin(baudrate, SERIAL_8N2)) {
    Serial.println("Failed to start Modbus RTU Client!");
    while (1)
        ;
  }
}

/**
  Obtains T5 data type variable

  @param dev_address Device address.
  @param base_reg Register address.
*/
float getT5(int dev_address, int base_reg) {
  ModbusRTUClient.requestFrom(dev_address, INPUT_REGISTERS, base_reg - 30000, 2);

  while(!ModbusRTUClient.available()) {}

  uint32_t rawreg = ModbusRTUClient.read() << 16 | ModbusRTUClient.read();
  int8_t reg_exp = ((uint8_t*)&rawreg)[3];
  uint32_t reg_base = rawreg & 0x00FFFFFF;
  float reg = reg_base * pow(10, reg_exp);
  
  return reg;
}

/**
  Obtains T6 data type variable

  @param dev_address Device address.
  @param base_reg Register address.
*/
float getT6(int dev_address, int base_reg) {
  ModbusRTUClient.requestFrom(dev_address, INPUT_REGISTERS, base_reg - 30000, 2);
  
  while(!ModbusRTUClient.available()) {}
  
  uint32_t rawreg = ModbusRTUClient.read() << 16 | ModbusRTUClient.read();

  int8_t reg_exp = ((uint8_t*)&rawreg)[3];
  int32_t reg_base = (int32_t)rawreg & 0x007FFFFF;
  if(rawreg & 0x800000) {
    reg_base = -reg_base;
  }
  float reg = reg_base * pow(10, reg_exp);

  return reg;
}

/**
  Obtains T2 data type variable

  @param dev_address Device address.
  @param base_reg Register address.
*/
float getT2(int dev_address, int base_reg) {
  ModbusRTUClient.requestFrom(dev_address, INPUT_REGISTERS, base_reg - 30000, 1);
  while(!ModbusRTUClient.available()) {}

  int16_t rawreg = ModbusRTUClient.read();
  
  return (float)rawreg;
}

/**
  Obtains T_Float data type variable

  @param dev_address Device address.
  @param base_reg Register address.
*/
float getT_Float(int dev_address, int base_reg) {
  ModbusRTUClient.requestFrom(dev_address, INPUT_REGISTERS, base_reg - 30000, 2);
  
  while(!ModbusRTUClient.available()) {}
  
  uint32_t rawreg = ModbusRTUClient.read() << 16 | ModbusRTUClient.read();
  float reg;
  memcpy(&reg, &rawreg, sizeof(float));
  
  return reg;
}


/*****************************
* Arduino Cloud 
*****************************/

/**
  Sets up configuration for Arduino Cloud
*/
void iot_cloud_setup(){
  // Defined in thingProperties.h
  initProperties();

  // Connect to Arduino IoT Cloud
  ArduinoCloud.begin(ArduinoIoTPreferredConnection);
  
  /*
     The following function allows you to obtain more information
     related to the state of network and IoT Cloud connection and errors
     the higher number the more granular information you’ll get.
     The default is 0 (only errors).
     Maximum is 4
 */
  setDebugMessageLevel(2);
  ArduinoCloud.printDebugInfo();
}

/*
  Since UOperMargin is READ_WRITE variable, onUOperMarginChange() is
  executed every time a new value is received from IoT Cloud.
*/
void onUOperMarginChange()  {
  Serial.println(F("Energy Manager: Operation margin updated from Arduino Cloud"));
  if (uOperMargin > 1){
    user_profile.uV_code = uOperMargin;
  }
}

/*
  Since UWatt is READ_WRITE variable, onUWattChange() is
  executed every time a new value is received from IoT Cloud.
*/
void onUWattChange()  {
  Serial.println(F("Energy Manager: Power value updated from Arduino Cloud"));
  if (uWatt > 1){
    user_profile.uW_code = uWatt;
  }
}

/*
  Since UWhCon is READ_WRITE variable, onUWhConChange() is
  executed every time a new value is received from IoT Cloud.
*/
void onUWhConChange()  {
  Serial.println(F("Energy Manager: Energy [Wh] value updated from Arduino Cloud"));
  if (uWhCon > 1){
    user_profile.uWh_code = uWhCon;
  }
}

/*
  Since DirectOverride1 is READ_WRITE variable, onDirectOverride1Change() is
  executed every time a new value is received from IoT Cloud.
*/
void onDirectOverride1Change()  {
  if (directOverride1 == true){
    Serial.println(F("Energy Manager: Device #1 Override ON from Arduino Cloud"));
    digitalWrite(D0, HIGH); 
  }
  if (directOverride1 == false){
    // D0 will operate depending on the system's resource state
    Serial.println(F("Energy Manager: Device #1 under system control"));
  } 
}

/*****************************
* Main
*****************************/
void setup() {
  // Initial Parameter 
  directOverride1 = false;
  uWhOpta = 0;

  Serial.begin(9600);
  delay(1000);

  // Analog/Digital IO Port Configuration
  analogIO_Setup();
  digitalIO_Setup();

  // Modbus RTU Configuration 
  RTU_Setup();
  
  // Status LED configuration;
  plc_led_Setup();

  // IoT Cloud Setup
  iot_cloud_setup();

  // Configures user profile and controlled device parameters
  consumption_profile(operation_safety_margin, estimated_max_power, estimated_max_energy);
  
  // Only for Device On State flag
  digitalWrite(LEDG, HIGH);

  // Scheduler -> ModBus
  Scheduler.startLoop(modbus_line);
}

void loop() {
  // Cloud exchange
  consumption_var_container();
  ArduinoCloud.update();

  // Profile based energy management tasks
  energy_distro_ctrl();

  // Serial print of the data from the Energy Meter
  modbus_com_monitor();

  delay(500);
}

/**
  Dedicated function for scheduler to retrieve Energy meter's information over Modbus RTU protocol
*/
void modbus_line(){
  modbus_com_actual();
  modbus_com_avg();
  modbus_com_max();
  modbus_com_min();
  modbus_com_energy();
}